



python -c "from __future__ import print_function; import conda; print(conda.__version__)"
IF %ERRORLEVEL% NEQ 0 exit 1
where conda
IF %ERRORLEVEL% NEQ 0 exit 1
where conda-env
IF %ERRORLEVEL% NEQ 0 exit 1
conda --version
IF %ERRORLEVEL% NEQ 0 exit 1
conda info
IF %ERRORLEVEL% NEQ 0 exit 1
conda env --help
IF %ERRORLEVEL% NEQ 0 exit 1
exit 0
